﻿// セキュア？ このプログラムを実行できる時点で手遅れだってば
#define _CRT_SECURE_NO_WARNINGS

#include <stdio.h>
#include <string>
#include <cstring>
using namespace std;
using std::string;

const string VER_NUM = "0.03";              // バージョン
const string COPYRIGHT = "(c) 2022 AKT";    // コピーライト

// データコピー用バッファサイズ
#define BUF_SIZE (1024*1000*16)

// サブルーチン宣言
extern bool getHdsInfo(FILE *src);          // HDS イメージ情報の取得
extern void getFatSize(size_t areaSize);    // FAT サイズの取得

// サブルーチンから受取用グローバル変数 (お行儀悪い)
size_t g_startPos = 0;                      // 領域開始オフセット
size_t g_fatSize = 0;                       // FAT サイズ
size_t g_rootPos = 0;                       // ルートディレクトリの相対位置
size_t g_partSize = 0;                      // ルートディレクトリ以降のデータ領域サイズ

// メインルーチン
int main(int argc, char* argv[])
{
    size_t n;               // 汎用読み込みサイズ一時記憶用
    FILE *src, *dst;        // 入出力ファイルポインタ
    unsigned char *buf;     // 汎用バッファ
    string srcName = "";    // 入力ファイル名
    string dstName = "";    // 出力ファイル名

    // タイトル表示
    fprintf(stdout, "X68k HDS to VDISK Converter v%s %s\n", VER_NUM.c_str(), COPYRIGHT.c_str());
    if (argc != 3)
    {
        // 使い方表示
        fprintf(stdout, "  [Usage] %s (HDS file) (VDISK file)\n", *argv);
    }
    else
    {
        srcName = *++argv;  //   HDS 入力ファイル名
        dstName = *++argv;  // VDISK 出力ファイル名

        // TODO : ホントはフルパス文字列化してから比較すべき
        if (_stricmp(srcName.c_str(), dstName.c_str()) == 0)
        {
            fprintf(stderr, "入出力に同じファイル %s は指定できません\n", srcName.c_str());
            return -1;
        }

        if ((src = fopen(srcName.c_str(), "rb")) == NULL)
        {
            fprintf(stderr, "ファイル %s をオープンできませんでした\n", srcName.c_str());
            return -1;
        }
        else if ((dst = fopen(dstName.c_str(), "wb")) == NULL)
        {
            fprintf(stderr, "ファイル %s をオープンできませんでした\n", dstName.c_str());
            fclose(src);
            return -1;
        }
        else
        {
            // HDS イメージ情報の取得
            if (getHdsInfo(src) == false)
            {
                fprintf(stderr, "HDS イメージの解析に失敗しました\n");
                fclose(src);
                fclose(dst);
                return -3;
            }

            fprintf(stdout, "[ CONVERT START ]\n");

            // FAT の開始場所
            fseek(src, (long)g_startPos + 0x400, SEEK_SET);
            fprintf(stdout, "IMAGE HEADER SKIP : %zd\n", g_startPos + 0x400);

            buf = (unsigned char*)malloc(g_fatSize);
            if (buf)
            {
                // ----- FAT 部分をコピーする
                // FAT 読み込み
                n = fread(buf, sizeof(unsigned char), g_fatSize, src);
                fprintf(stdout, "         FAT COPY : %zd\n", n);

                // FAT 先頭改ざん (VDISK に見せる)
                buf[0] = 0xF9;
                // FAT 書き込み
                fwrite(buf, sizeof(unsigned char), n, dst);
                free(buf);

                // ルートディレクトリの場所までジャンプ
                fseek(src, (long)g_startPos + 0x400 + (long)g_rootPos, SEEK_SET);
                fprintf(stdout, "    ROOT POSITION : %zd\n", g_startPos + 0x400 + g_rootPos);

                // ----- データ領域をコピーする
                // VDISK では FAT 直後にルートフォルダ領域を置く
                fprintf(stdout, "  DATA COPY START : ");
                buf = (unsigned char*)malloc(BUF_SIZE);
                if (buf)
                {
                    size_t write_size = g_partSize;  // 書き出す残りファイルサイズ
                    // 中身コピー
                    while (write_size > 0)
                    {
                        n = fread(buf, sizeof(unsigned char), (write_size > BUF_SIZE) ? BUF_SIZE : write_size, src);

                        // TODO : n を見てエラーチェックしようね

                        fprintf(stdout, "*");
                        fwrite(buf, sizeof(unsigned char), n, dst);
                        write_size -= n;
                    }
                    fprintf(stdout, "\n");
                    fprintf(stdout, "    DATA COPY END : %zd\n", g_partSize);
                    fprintf(stdout, "多分完了しました\n");
                    free(buf);
                }
                else
                {
                    fprintf(stderr, "メモリ確保に失敗しました 2\n");
                    return -2;
                }
            }
            else
            {
                fprintf(stderr, "メモリ確保に失敗しました\n");
                return -2;
            }
            fclose(src);
            fclose(dst);
        }
    }
    return 0;
}

// 16bit 取得
size_t get16bit(unsigned char* b)
{
    return ((size_t)b[0] << 8) + (size_t)b[1];
}

// 24bit 取得
size_t get24bit(unsigned char* b)
{
    return ((size_t)b[0] << 16) + ((size_t)b[1] << 8) + (size_t)b[2];
}

// 32bit 取得
size_t get32bit(unsigned char* b)
{
    return ((size_t)b[0] << 24) + ((size_t)b[1] << 16) + ((size_t)b[2] << 8) + (size_t)b[3];
}

// HDS イメージ情報の取得
bool getHdsInfo(FILE *src)
{
    unsigned char* buf;         // 汎用バッファ
    int HEADER_SIZE = 0x20000;  // ちょっと多めに最初のパーティション IPL まで読み込む

    // TODO : パーティション全部抜き出すなら真面目に分割して読まないと
    //        というか書き出し部分の方をサブルーチン化して構成を変えないとね

    // ヘッダ読み込み
    buf = (unsigned char*)malloc(HEADER_SIZE);
    if (buf)
    {
        size_t n = fread(buf, sizeof(unsigned char), HEADER_SIZE, src);
        // マジック文字列チェック
        if ( (n != HEADER_SIZE) ||
            (buf[0] != 'X') || (buf[1] != '6') || (buf[2] != '8') || (buf[3] != 'S') ||
            (buf[4] != 'C') || (buf[5] != 'S') || (buf[6] != 'I') || (buf[7] != '1'))
        {
            fprintf(stderr, "HDS ヘッダが見つかりません\n");
            free(buf);
            return false;
        }

        fprintf(stdout, "[ .HDS FILE INFORMATION ]\n");

        size_t byteNum = get16bit(&buf[8]);                                 // １レコード当たりのバイト数
        fprintf(stdout, "  BYTE per RECORD : %zd\n", byteNum);
        size_t recordNum = get32bit(&buf[10]);                              // レコード数
        fprintf(stdout, "       RECORD NUM : %zd\n", recordNum);
        size_t allSize = (byteNum * recordNum) + 512;                       // イメージサイズ (+512 は謎(ぉ)
        fprintf(stdout, "    ALL DISK SIZE : %zd (%zdMB)\n", allSize, (allSize / 1024 / 1024));

        // パーティションテーブルチェック
        for (int p = 1; p < 9; p++)
        {
            char* pName = (char*)&buf[0x800 + (p * 0x10)];
            string partitionName = "";                                      // [8B] パーティション名
            for (int i = 0; i < 8; i++)
            {
                partitionName += *pName++;
            }
            partitionName += '\0';
                                                                            // [1B] 種類とかあったけどシカト
            size_t offsetNum = get24bit(&buf[0x809 + (p * 0x10)]) * 1024;   // [3B] 領域開始オフセット (IPLあり。FAT開始位置は+0x400)
            size_t sizeNum = get32bit(&buf[0x80C + (p * 0x10)]) * 1024;     // [4B] 領域サイズ
            if ((offsetNum == 0) && (sizeNum == 0))
            {
                break;                                                      // パーティションが存在しない
            }
            else
            {
                size_t rootPos = get16bit(&buf[offsetNum + 0x1D]) * 8;      // ルートディレクトリの相対位置
                g_startPos = offsetNum;                                     // 領域開始オフセットを受け渡し用変数へ
                g_rootPos = rootPos;                                        // ルートディレクトリの相対位置を受け渡し用変数へ
                sizeNum -= (g_rootPos + 0x400);                             // ルートディレクトリ以降のデータ領域サイズ
                g_partSize = sizeNum;                                       // ルートディレクトリ以降のデータ領域サイズを受け渡し用変数へ

                fprintf(stdout, "[ PARTITION %d ]\n", p);
                fprintf(stdout, "             NAME : %s\n", partitionName.c_str());
                fprintf(stdout, "        START POS : %zd (0x%zX)\n", offsetNum, offsetNum);
                fprintf(stdout, "         ROOT POS : %zd (0x%zX)\n", rootPos, rootPos);
                fprintf(stdout, "        DATA SIZE : %zd (%zdMB)\n", sizeNum, (sizeNum / 1024 / 1024));

                // FAT サイズの取得
                getFatSize(sizeNum);

                break;                                                      // 今は最初のパーティションだけ
            }
        }
        free(buf);
    }
    else
    {
        fprintf(stderr, "ヘッダ読み込みのメモリが確保出来ません\n");
        return false;
    }
    return true;
}

// FAT サイズの取得
void getFatSize(size_t areaSize)
{
    // FAT サイズの算出開始
    fprintf(stdout, "[ FAT INFORMATION ]\n");

    // TODO : 本当はココで出た FATサイズ をデータ領域へ足して再計算しないとダメな気もするが一旦様子見(ぉ

    // 各マジックナンバーは考えちゃイケない、感じる
    size_t cluster_size = 1024;
    size_t tmpSize = areaSize;

    tmpSize /= cluster_size;
    tmpSize -= 0x10;

    while (0xFFF0 < tmpSize)
    {
        tmpSize /= 2;
        cluster_size *= 2;
    }
    fprintf(stdout, "     CLUSTER SIZE : %zd\n", cluster_size);

    tmpSize += 2;
    tmpSize *= 2;
    tmpSize += 0x3FF;
    tmpSize /= 0x400;

    size_t fat_SIZE = tmpSize * 1024;
    g_fatSize = fat_SIZE;
    fprintf(stdout, "         FAT SIZE : %zd\n", fat_SIZE);
}
